@@ -48,7 +48,6 @@ from xonsh.platform import (
     DEFAULT_ENCODING,
     ON_LINUX,
     ON_WINDOWS,
-    PYTHON_VERSION_INFO,
     expanduser,
     os_environ,
     pygments_version_info,
@@ -78,6 +77,7 @@ class XonshCalledProcessError(XonshError, subprocess.CalledProcessError):
     returncode of the command is nonzero.
 
     Example:
+    -------
         try:
             for line in !(ls):
                 print(line)
@@ -516,9 +516,9 @@ def _have_open_triple_quotes(s):
 
 
 def get_line_continuation():
-    """ The line continuation characters used in subproc mode. In interactive
-         mode on Windows the backslash must be preceded by a space. This is because
-         paths on Windows may end in a backslash.
+    """The line continuation characters used in subproc mode. In interactive
+    mode on Windows the backslash must be preceded by a space. This is because
+    paths on Windows may end in a backslash.
     """
     if (
         ON_WINDOWS
@@ -692,7 +692,7 @@ def indent(instr, nspaces=4, ntabs=0, flatten=False):
 
 
 def get_sep():
-    """ Returns the appropriate filepath separator char depending on OS and
+    """Returns the appropriate filepath separator char depending on OS and
     xonsh options set
     """
     if ON_WINDOWS and builtins.__xonsh__.env.get("FORCE_POSIX_PATHS"):
@@ -775,15 +775,6 @@ def _yield_accessible_unix_file_names(path):
 def _executables_in_posix(path):
     if not os.path.exists(path):
         return
-    elif PYTHON_VERSION_INFO < (3, 5, 0):
-        for fname in os.listdir(path):
-            fpath = os.path.join(path, fname)
-            if (
-                os.path.exists(fpath)
-                and os.access(fpath, os.X_OK)
-                and (not os.path.isdir(fpath))
-            ):
-                yield fname
     else:
         yield from _yield_accessible_unix_file_names(path)
 
@@ -792,32 +783,24 @@ def _executables_in_windows(path):
     if not os.path.isdir(path):
         return
     extensions = builtins.__xonsh__.env["PATHEXT"]
-    if PYTHON_VERSION_INFO < (3, 5, 0):
-        for fname in os.listdir(path):
-            fpath = os.path.join(path, fname)
-            if os.path.exists(fpath) and not os.path.isdir(fpath):
-                base_name, ext = os.path.splitext(fname)
-                if ext.upper() in extensions:
-                    yield fname
-    else:
-        try:
-            for x in scandir(path):
-                try:
-                    is_file = x.is_file()
-                except OSError:
-                    continue
-                if is_file:
-                    fname = x.name
-                else:
-                    continue
-                base_name, ext = os.path.splitext(fname)
-                if ext.upper() in extensions:
-                    yield fname
-        except FileNotFoundError:
-            # On Windows, there's no guarantee for the directory to really
-            # exist even if isdir returns True. This may happen for instance
-            # if the path contains trailing spaces.
-            return
+    try:
+        for x in scandir(path):
+            try:
+                is_file = x.is_file()
+            except OSError:
+                continue
+            if is_file:
+                fname = x.name
+            else:
+                continue
+            base_name, ext = os.path.splitext(fname)
+            if ext.upper() in extensions:
+                yield fname
+    except FileNotFoundError:
+        # On Windows, there's no guarantee for the directory to really
+        # exist even if isdir returns True. This may happen for instance
+        # if the path contains trailing spaces.
+        return
 
 
 def executables_in(path):
@@ -832,7 +815,7 @@ def executables_in(path):
         return
 
 
-def command_not_found(cmd):
+def debian_command_not_found(cmd):
     """Uses the debian/ubuntu command-not-found utility to suggest packages for a
     command that cannot currently be found.
     """
@@ -857,6 +840,33 @@ def command_not_found(cmd):
     return s
 
 
+def conda_suggest_command_not_found(cmd, env):
+    """Uses conda-suggest to suggest packages for a command that cannot
+    currently be found.
+    """
+    try:
+        from conda_suggest import find
+    except ImportError:
+        return ""
+    return find.message_string(
+        cmd, conda_suggest_path=env.get("CONDA_SUGGEST_PATH", None)
+    )
+
+
+def command_not_found(cmd, env):
+    """Uses various mechanism to suggest packages for a command that cannot
+    currently be found.
+    """
+    if ON_LINUX:
+        rtn = debian_command_not_found(cmd)
+    else:
+        rtn = ""
+    conda = conda_suggest_command_not_found(cmd, env)
+    if conda:
+        rtn = rtn + "\n\n" + conda if rtn else conda
+    return rtn
+
+
 def suggest_commands(cmd, env, aliases):
     """Suggests alternative commands given an environment and aliases."""
     if not env.get("SUGGEST_COMMANDS"):
@@ -873,13 +883,10 @@ def suggest_commands(cmd, env, aliases):
             if levenshtein(alias.lower(), cmd, thresh) < thresh:
                 suggested[alias] = "Alias"
 
-    for path in filter(os.path.isdir, env.get("PATH")):
-        for _file in executables_in(path):
-            if (
-                _file not in suggested
-                and levenshtein(_file.lower(), cmd, thresh) < thresh
-            ):
-                suggested[_file] = "Command ({0})".format(os.path.join(path, _file))
+    for _cmd in builtins.__xonsh__.commands_cache.all_commands:
+        if _cmd not in suggested:
+            if levenshtein(_cmd.lower(), cmd, thresh) < thresh:
+                suggested[_cmd] = "Command ({0})".format(_cmd)
 
     suggested = collections.OrderedDict(
         sorted(
@@ -889,7 +896,7 @@ def suggest_commands(cmd, env, aliases):
     num = min(len(suggested), max_sugg)
 
     if num == 0:
-        rtn = command_not_found(cmd)
+        rtn = command_not_found(cmd, env)
     else:
         oneof = "" if num == 1 else "one of "
         tips = "Did you mean {}the following?".format(oneof)
@@ -899,30 +906,71 @@ def suggest_commands(cmd, env, aliases):
             "    {: <{}} {}".format(key + ":", length, val) for key, val in items
         )
         rtn = "{}\n{}".format(tips, alternatives)
-        c = command_not_found(cmd)
+        c = command_not_found(cmd, env)
         rtn += ("\n\n" + c) if len(c) > 0 else ""
     return rtn
 
 
-def print_exception(msg=None):
-    """Print exceptions with/without traceback."""
+def _get_manual_env_var(name, default=None):
+    """Returns if the given variable is manually set as well as it's value."""
     env = getattr(builtins.__xonsh__, "env", None)
-    # flags indicating whether the traceback options have been manually set
     if env is None:
         env = os_environ
-        manually_set_trace = "XONSH_SHOW_TRACEBACK" in env
-        manually_set_logfile = "XONSH_TRACEBACK_LOGFILE" in env
+        manually_set = name in env
     else:
-        manually_set_trace = env.is_manually_set("XONSH_SHOW_TRACEBACK")
-        manually_set_logfile = env.is_manually_set("XONSH_TRACEBACK_LOGFILE")
+        manually_set = env.is_manually_set(name)
+
+    value = env.get(name, default)
+    return (manually_set, value)
+
+
+def print_warning(msg):
+    """Print warnings with/without traceback."""
+    manually_set_trace, show_trace = _get_manual_env_var("XONSH_SHOW_TRACEBACK", False)
+    manually_set_logfile, log_file = _get_manual_env_var("XONSH_TRACEBACK_LOGFILE")
     if (not manually_set_trace) and (not manually_set_logfile):
         # Notify about the traceback output possibility if neither of
         # the two options have been manually set
         sys.stderr.write(
             "xonsh: For full traceback set: " "$XONSH_SHOW_TRACEBACK = True\n"
         )
-    # get env option for traceback and convert it if necessary
-    show_trace = env.get("XONSH_SHOW_TRACEBACK", False)
+    # convert show_trace to bool if necessary
+    if not is_bool(show_trace):
+        show_trace = to_bool(show_trace)
+    # if the trace option has been set, print all traceback info to stderr
+    if show_trace:
+        # notify user about XONSH_TRACEBACK_LOGFILE if it has
+        # not been set manually
+        if not manually_set_logfile:
+            sys.stderr.write(
+                "xonsh: To log full traceback to a file set: "
+                "$XONSH_TRACEBACK_LOGFILE = <filename>\n"
+            )
+        traceback.print_stack()
+    # additionally, check if a file for traceback logging has been
+    # specified and convert to a proper option if needed
+    log_file = to_logfile_opt(log_file)
+    if log_file:
+        # if log_file <> '' or log_file <> None, append
+        # traceback log there as well
+        with open(os.path.abspath(log_file), "a") as f:
+            traceback.print_stack(file=f)
+
+    msg = msg if msg.endswith("\n") else msg + "\n"
+    sys.stderr.write(msg)
+
+
+def print_exception(msg=None):
+    """Print exceptions with/without traceback."""
+    manually_set_trace, show_trace = _get_manual_env_var("XONSH_SHOW_TRACEBACK", False)
+    manually_set_logfile, log_file = _get_manual_env_var("XONSH_TRACEBACK_LOGFILE")
+    if (not manually_set_trace) and (not manually_set_logfile):
+        # Notify about the traceback output possibility if neither of
+        # the two options have been manually set
+        sys.stderr.write(
+            "xonsh: For full traceback set: " "$XONSH_SHOW_TRACEBACK = True\n"
+        )
+    # convert show_trace to bool if necessary
     if not is_bool(show_trace):
         show_trace = to_bool(show_trace)
     # if the trace option has been set, print all traceback info to stderr
@@ -937,7 +985,6 @@ def print_exception(msg=None):
         traceback.print_exc()
     # additionally, check if a file for traceback logging has been
     # specified and convert to a proper option if needed
-    log_file = env.get("XONSH_TRACEBACK_LOGFILE", None)
     log_file = to_logfile_opt(log_file)
     if log_file:
         # if log_file <> '' or log_file <> None, append
@@ -1029,7 +1076,7 @@ def escape_windows_cmd_string(s):
 
 
 def argvquote(arg, force=False):
-    """ Returns an argument quoted in such a way that that CommandLineToArgvW
+    """Returns an argument quoted in such a way that that CommandLineToArgvW
     on Windows will return the argument string unchanged.
     This is the same thing Popen does when supplied with an list of arguments.
     Arguments in a command line should be separated by spaces; this
@@ -1164,11 +1211,33 @@ def ensure_string(x):
     return str(x)
 
 
+def is_path(x):
+    """This tests if something is a path."""
+    return isinstance(x, pathlib.Path)
+
+
 def is_env_path(x):
     """This tests if something is an environment path, ie a list of strings."""
     return isinstance(x, EnvPath)
 
 
+def str_to_path(x):
+    """Converts a string to a path."""
+    if x is None:
+        return None
+    elif isinstance(x, str):
+        # checking x is needed to avoid uncontrolled converting empty string to Path('.')
+        return pathlib.Path(x) if x else None
+    elif isinstance(x, pathlib.Path):
+        return x
+    elif isinstance(x, EnvPath) and len(x) == 1:
+        return pathlib.Path(x[0]) if x[0] else None
+    else:
+        raise TypeError(
+            f"Variable should be a pathlib.Path, str or single EnvPath type. {type(x)} given."
+        )
+
+
 def str_to_env_path(x):
     """Converts a string to an environment path, ie a list of strings,
     splitting on the OS separator.
@@ -1177,6 +1246,11 @@ def str_to_env_path(x):
     return EnvPath(x)
 
 
+def path_to_str(x):
+    """Converts a path to a string."""
+    return str(x)
+
+
 def env_path_to_str(x):
     """Converts an environment path to a string by joining on the OS
     separator.
@@ -1189,6 +1263,11 @@ def is_bool(x):
     return isinstance(x, bool)
 
 
+def is_bool_or_none(x):
+    """Tests if something is a boolean or None."""
+    return (x is None) or isinstance(x, bool)
+
+
 def is_logfile_opt(x):
     """
     Checks if x is a valid $XONSH_TRACEBACK_LOGFILE option. Returns False
@@ -1208,10 +1287,7 @@ def to_logfile_opt(x):
     the filepath if it is a writable file or None if the filepath is not
     valid, informing the user on stderr about the invalid choice.
     """
-    superclass = pathlib.PurePath
-    if PYTHON_VERSION_INFO >= (3, 6, 0):
-        superclass = os.PathLike
-    if isinstance(x, superclass):
+    if isinstance(x, os.PathLike):
         x = str(x)
     if is_logfile_opt(x):
         return x
@@ -1246,7 +1322,7 @@ _FALSES = LazyObject(
 
 
 def to_bool(x):
-    """"Converts to a boolean in a semantically meaningful way."""
+    """Converts to a boolean in a semantically meaningful way."""
     if isinstance(x, bool):
         return x
     elif isinstance(x, str):
@@ -1255,6 +1331,20 @@ def to_bool(x):
         return bool(x)
 
 
+def to_bool_or_none(x):
+    """Converts to a boolean or none in a semantically meaningful way."""
+    if x is None or isinstance(x, bool):
+        return x
+    elif isinstance(x, str):
+        low_x = x.lower()
+        if low_x == "none":
+            return None
+        else:
+            return False if x.lower() in _FALSES else True
+    else:
+        return bool(x)
+
+
 def to_itself(x):
     """No conversion, returns itself."""
     return x
@@ -1267,6 +1357,14 @@ def bool_to_str(x):
     return "1" if x else ""
 
 
+def bool_or_none_to_str(x):
+    """Converts a bool or None value to a string."""
+    if x is None:
+        return "None"
+    else:
+        return "1" if x else ""
+
+
 _BREAKS = LazyObject(
     lambda: frozenset(["b", "break", "s", "skip", "q", "quit"]), globals(), "_BREAKS"
 )
@@ -1475,8 +1573,8 @@ def bool_seq_to_csv(x):
 
 
 def ptk2_color_depth_setter(x):
-    """ Setter function for $PROMPT_TOOLKIT_COLOR_DEPTH. Also
-        updates os.environ so prompt toolkit can pickup the value.
+    """Setter function for $PROMPT_TOOLKIT_COLOR_DEPTH. Also
+    updates os.environ so prompt toolkit can pickup the value.
     """
     x = str(x)
     if x in {
@@ -1654,7 +1752,7 @@ def is_history_backend(x):
 
 
 def is_dynamic_cwd_width(x):
-    """ Determine if the input is a valid input for the DYNAMIC_CWD_WIDTH
+    """Determine if the input is a valid input for the DYNAMIC_CWD_WIDTH
     environment variable.
     """
     return (
@@ -1866,7 +1964,7 @@ WIN_BOLD_COLOR_MAP = LazyObject(_win_bold_color_map, globals(), "WIN_BOLD_COLOR_
 
 def hardcode_colors_for_win10(style_map):
     """Replace all ansi colors with hardcoded colors to avoid unreadable defaults
-       in conhost.exe
+    in conhost.exe
     """
     modified_style = {}
     if not builtins.__xonsh__.env["PROMPT_TOOLKIT_COLOR_DEPTH"]:
@@ -1892,8 +1990,7 @@ def hardcode_colors_for_win10(style_map):
 
 
 def ansicolors_to_ptk1_names(stylemap):
-    """Converts ansicolor names in a stylemap to old PTK1 color names
-    """
+    """Converts ansicolor names in a stylemap to old PTK1 color names"""
     if pygments_version_info() and pygments_version_info() >= (2, 4, 0):
         return stylemap
     modified_stylemap = {}
@@ -1907,7 +2004,7 @@ def ansicolors_to_ptk1_names(stylemap):
 
 def intensify_colors_for_cmd_exe(style_map):
     """Returns a modified style to where colors that maps to dark
-       colors are replaced with brighter versions.
+    colors are replaced with brighter versions.
     """
     modified_style = {}
     replace_colors = {
@@ -2143,14 +2240,18 @@ def expandvars(path):
         # get the path's string representation
         path = str(path)
     if "$" in path:
+        shift = 0
         for match in POSIX_ENVVAR_REGEX.finditer(path):
             name = match.group("envvar")
             if name in env:
-                ensurer = env.get_ensurer(name)
+                detyper = env.get_detyper(name)
                 val = env[name]
-                value = str(val) if ensurer.detype is None else ensurer.detype(val)
+                value = str(val) if detyper is None else detyper(val)
                 value = str(val) if value is None else value
-                path = POSIX_ENVVAR_REGEX.sub(value, path, count=1)
+                start_pos, end_pos = match.span()
+                path_len_before_replace = len(path)
+                path = path[: start_pos + shift] + value + path[end_pos + shift :]
+                shift = shift + len(path) - path_len_before_replace
     return path
 
 
@@ -2258,37 +2359,21 @@ def _iglobpath(s, ignore_case=False, sort_result=None, include_dotfiles=None):
         include_dotfiles = builtins.__xonsh__.env.get("DOTGLOB")
     if ignore_case:
         s = expand_case_matching(s)
-    if sys.version_info > (3, 5):
-        if "**" in s and "**/*" not in s:
-            s = s.replace("**", "**/*")
-        if include_dotfiles:
-            dotted_s, dotmodified = _dotglobstr(s)
-        # `recursive` is only a 3.5+ kwarg.
-        if sort_result:
-            paths = glob.glob(s, recursive=True)
-            if include_dotfiles and dotmodified:
-                paths.extend(glob.iglob(dotted_s, recursive=True))
-            paths.sort()
-            paths = iter(paths)
-        else:
-            paths = glob.iglob(s, recursive=True)
-            if include_dotfiles and dotmodified:
-                paths = itertools.chain(glob.iglob(dotted_s, recursive=True), paths)
-        return paths, s
+    if "**" in s and "**/*" not in s:
+        s = s.replace("**", "**/*")
+    if include_dotfiles:
+        dotted_s, dotmodified = _dotglobstr(s)
+    if sort_result:
+        paths = glob.glob(s, recursive=True)
+        if include_dotfiles and dotmodified:
+            paths.extend(glob.iglob(dotted_s, recursive=True))
+        paths.sort()
+        paths = iter(paths)
     else:
-        if include_dotfiles:
-            dotted_s, dotmodified = _dotglobstr(s)
-        if sort_result:
-            paths = glob.glob(s)
-            if include_dotfiles and dotmodified:
-                paths.extend(glob.iglob(dotted_s))
-            paths.sort()
-            paths = iter(paths)
-        else:
-            paths = glob.iglob(s)
-            if include_dotfiles and dotmodified:
-                paths = itertools.chain(glob.iglob(dotted_s), paths)
-        return paths, s
+        paths = glob.iglob(s, recursive=True)
+        if include_dotfiles and dotmodified:
+            paths = itertools.chain(glob.iglob(dotted_s, recursive=True), paths)
+    return paths, s
 
 
 def iglobpath(s, ignore_case=False, sort_result=None, include_dotfiles=None):
